/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Communicator client code, released
 * March 31, 1998.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Jeff Hammel <jhammel@mozilla.com>
 *    Jonathan Griffin <jgriffin@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// (generic) utility functions used by ProfileManager

Components.utils.import("resource://gre/modules/ctypes.jsm");

const Ci = Components.interfaces;
const Cc = Components.classes;
const CC = Components.Constructor;

// some XPCOM services
const dnsService = Cc["@mozilla.org/network/dns-service;1"].
                   getService(Ci.nsIDNSService);
const envService = Cc["@mozilla.org/process/environment;1"].
                   getService(Ci.nsIEnvironment);
const prefService = Cc["@mozilla.org/preferences-service;1"].
                    getService(Ci.nsIPrefBranch).
                    QueryInterface(Ci.nsIPrefService);
const promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"].
                      getService(Ci.nsIPromptService);
const directoryService = Cc["@mozilla.org/file/directory_service;1"].
                         getService(Ci.nsIProperties);
const nsIFilePicker = Ci.nsIFilePicker;

function getFOStream() {
  return Cc["@mozilla.org/network/file-output-stream;1"].
         createInstance(Ci.nsIFileOutputStream);
}
function getConverterOS() {
  return Cc["@mozilla.org/intl/converter-output-stream;1"].
         createInstance(Ci.nsIConverterOutputStream);
}
function getFilePicker() {
  return Cc["@mozilla.org/filepicker;1"].
         createInstance(nsIFilePicker);
}

// get our IP address
var gPlatform = Cc["@mozilla.org/network/protocol;1?name=http"].
                getService(Ci.nsIHttpProtocolHandler).platform;
var gMyIP = "127.0.0.1";
try {
  gMyIP = dnsService.resolve(dnsService.myHostName, 0).getNextAddrAsString();
}
catch(e) { }

/**
 * Determines if the given symlink, associated with a profile lock,
 * is stale.
 * @param file {nsIFile} the symlink to test; this is guaranteed to exist
 *        and be a symlink
 * @return true if the symlink is stale (the lock is invalid), otherwise
 *         false (the lock is valid)
 */
function isSymlinkStale(file) {
  // see http://mxr.mozilla.org/mozilla-central/source/profile/dirserviceprovider/src/nsProfileLock.cpp#279

  var ret = ctypes.int.ptr;

  // OS-specific declarations
  if (gPlatform == "Macintosh") {
    var libc = ctypes.open("libc.dylib");
    // this gives as access to errno on Mac
    var error = libc.declare("__error",
                              ctypes.default_abi,
                              ret);
  }
  else if (gPlatform == "X11") {
    var libc = ctypes.open("libc.so.6");
    // this gives us access to errno on Linux
    var error = libc.declare("__errno_location",
                             ctypes.default_abi,
                             ret);
  }
  else {
    throw document.getElementById("bundle_profileManager").
                   getFormattedString('isSymlinkStaleNotSupported', 
                                      [gPlatform]);
  }

  // string buffer that readlink can use to copy data into
  var bufstr = "                                                     ";
  var buffer = ctypes.unsigned_char.array()(bufstr);

  // readlink:
  // http://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man2/readlink.2.html
  var readlink = libc.declare("readlink",
                              ctypes.default_abi,
                              ctypes.ssize_t,
                              ctypes.unsigned_char.ptr,
                              ctypes.unsigned_char.ptr,
                              ctypes.size_t);

  // kill:
  // http://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man2/kill.2.html
  var kill = libc.declare("kill",
                          ctypes.default_abi,
                          ctypes.int,
                          ctypes.int32_t,
                          ctypes.int);

  // read the contents of the symlink; if empty, return true
  var read = readlink(file.path, buffer, bufstr.length);
  if (read <= 0) 
    return true;

  // the contents of the symlink should be IP:[+]pid; if not return true
  var output = buffer.readString().trim();
  var parts = output.split(":");
  if (parts.length < 2)
    return true;
  let ip = parts[0];
  let pid = parts[1];

  // if pid starts with '+', it indicates a fcntl lock; strip it and continue
  if (pid[0] == "+")
    pid = pid.substr(1);

  // if the ip address in the symlink doesn't match the current ip, it means
  // the symlink was created on a different machine, so return false
  if (ip != gMyIP) {
    return false;
  }

  // attempt to see if the pid is real by using kill(pid, 0)
  let killreturn = kill(Number(pid), 0);
  let errno = error().contents;
  if (killreturn == 0 || errno == 3) { // ESRCH 3 No such process
    // the process that created the symlink still exists, return false
    return false;
  }
  libc.close();

  // by default return true
  return true;
}

/**
 * Determines if a profile lockfile is locked with fcntl.  This is done by
 * attempting to lock the file; if we can, it isn't locked by something else.
 * @param file {nsIFile} the file to check
 * @return true if the file is already fcntl locked, otherwise false
 */
function isFcntlLocked(file) {
  // check if the file exists
  if (!file.exists()) {
    return false;
  }
  var result = false;

  // OS-specific declarations
  var ret = ctypes.int.ptr;
  if (gPlatform == "Macintosh") {
    var libc = ctypes.open("libc.dylib");
    // this gives us access to errno on the Mac
    var error = libc.declare("__error",
                              ctypes.default_abi,
                              ret);
  }
  else if (gPlatform == "X11") {
    var libc = ctypes.open("libc.so.6");
    // this gives us access to errno on Linux
    var error = libc.declare("__errno_location",
                             ctypes.default_abi,
                             ret);
  }
  else {
    throw document.getElementById("bundle_profileManager").
                   getFormattedString('isFcntlLockedNotSupported',
                                      [gPlatform]);
  }

  // flock and fcntl:
  // http://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man2/fcntl.2.html
  var flock = new ctypes.StructType("flock", [ {"l_start": ctypes.uint64_t},
                                               {"l_len": ctypes.uint64_t},
                                               {"l_pid": ctypes.int32_t},
                                               {"l_type": ctypes.short},
                                               {"l_whence": ctypes.short} ]);
  var fcntl = libc.declare("fcntl",
                            ctypes.default_abi,
                            ctypes.int,
                            ctypes.int,
                            ctypes.int,
                            flock.ptr);

  // open:
  // http://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man2/open.2.html
  var openc = libc.declare("open",
                            ctypes.default_abi,
                            ctypes.int,
                            ctypes.unsigned_char.ptr,
                            ctypes.int,
                            ctypes.uint16_t);

  // close:
  // http://developer.apple.com/library/mac/#documentation/Darwin/Reference/ManPages/man2/close.2.html
  var closec = libc.declare("close",
                            ctypes.default_abi,
                            ctypes.void_t,
                            ctypes.int);

  // see http://mxr.mozilla.org/mozilla-central/source/profile/dirserviceprovider/src/nsProfileLock.cpp#229
  // open the file with FREAD | O_CREAT | O_TRUNC
  var fp = openc(file.path, 0x01 | 0x0200 | 0x0400, 0666);
  var lockdata = new flock;
  lockdata.l_start = ctypes.UInt64(0);
  lockdata.l_len = ctypes.UInt64(0);
  lockdata.l_type = 3; // F_WRLCK
  lockdata.l_whence = 0; // SEEK_SET

  // attempt to get lock status; if this returns -1 it means fcntl isn't 
  // supported on this file
  var getlock = fcntl(fp, 7, lockdata.address()); // F_GETLK
  if (getlock != -1) {
    // attempt to lock the file
    var setlock = fcntl(fp, 8, lockdata.address()); // F_SETLK
    if (setlock == -1) {
      // if attempting to set the lock failed with specific error codes,
      // the file is already locked
      let errno = error().contents;
      if (errno == 35 || errno == 13) { // EAGAIN || EACCES
        result = true;
      }
    }
  }

  closec(fp);
  libc.close();
  return result;
}

/**
 * Returns true if a file with the given path exists.
 */
function fileExists(path) {
  let file = getFile(path);
  return file.exists();
}

/**
 * Returns an nsIFile object based on the given path.  This function
 * does not check for file existence.
 */
function getFile(path) {
  // returns file object from path
  var file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
  file.initWithPath(path);
  return file;
}

/**
 * Returns the value of an environment variable.
 */
function getEnv(variable) {
  return envService.get(variable);
}

/**
 * Emulates the posix 'which' command.
 * @param filename {string} the name of the executable to search the PATH for;
 *        this does not include OS-specific suffixes like .exe
 * @return an nsIFile of the full path to the file, or undefined
 */
function which(filename) {
  var path = getEnv('PATH');

  if (gPlatform == "Windows") {
    path = path.split(';');
    filename += ".exe";
  }
  else {
    path = path.split(':');
  }

  for (var i in path) {
    if (path[i] == ".")
      continue;
    var file = getFile(path[i]);
    file.append(filename);
    if (file.exists()) {
      // XXX should check to ensure it's executable?
      return file;
    }
  }
}

/**
 * Returns the version of a xulrunner app by looking for the 'Version'
 * section of it's corresponding application.ini.
 * @param filename {string} the full path of the application's executable
 * @return a string containing the app's version, or '?' if no version could
 *         be located
 */ 
function getAppVersion(filename) {
  try {
    var file = getFile(filename);
  }
  catch (e) {
    return '?';
  }
  try {
    // follow symlinks
    let target = file.target;
    file = getFile(target);
  }
  catch (e) { /* no-op */ }
  file = file.parent;
  file.append("application.ini");
  if (file.exists()) {
    try {
      let parser = getINIParser(file);
      let appData = read_section(parser, "App");
      if (appData && appData['Version'])
        return appData['Version'];
    }
    catch(e) {
      /* no-op */
    }
  }
  // Return ? as a default value.  This will occur e.g., in very old
  // versions of Firefox (1.5), which don't seem to use an application.ini.
  return "?";
}

/**
 * On Windows, attempt to locate an application by searching for it
 * in possible "Program Files" variants.
 * @param filename {string} the executable filename, not including extension
 * @param default_program_subdir {string} the default subdirectory under
 *        "Program Files", where this program lives
 * @return an nsIFile representing the executable, or undefined
 */
function findInProgramFiles(filename, default_program_subdir) {
  if (gPlatform == "Windows" && typeof(default_program_subdir) == "string") {
    filename += ".exe";
    var envVars = [
      "PROGRAMFILES",
      "ProgramFiles",
      "PROGRAMFILES(X86)",
      "ProgramFiles(x86)",
      "PROGRAMW6432",
      "ProgramW6432"
    ];
    for each (let envVar in envVars) {
      var ProgramDir = envService.get(envVar)
      if (ProgramDir) {
        var file = getFile(ProgramDir);
        file.append(default_program_subdir);
        file.append(filename);
        if (file.exists()) {
          return file;
        }
      }
    }
  }
}

/**
 * Attempt to find the default version of an application: first, search
 * the path, then (on Windows & Mac) search OS-standard locations.
 * @param app {string} the executable's filename, without extension
 * @param appProperties an object containing properties for this app;
 *        see ProfileManager._appData.
 * @return an nsIFile of the default application, or undefined
 */
function findDefaultApp(app, appProperties) {
  let app_default = which(app);
  if (app_default == undefined && appProperties) {
    if (appProperties.default_subdir && (gPlatform == "Windows")) {
      app_default = findInProgramFiles(app, appProperties.default_subdir);
    }
    else if (appProperties.name && (gPlatform == "Macintosh")) {
      let macapp = "/Applications/" + appProperties.name + 
                   ".app/Contents/MacOS/" + app + "-bin";
      let macfile = getFile(macapp);
      if (macfile && macfile.exists()) {
        app_default = macfile;
      }
    }
  }
  return app_default;
}

/**
 * Remove all children from an element.
 */
function removeChildren(element) {
  while (element.childNodes.length) {
    element.removeChild(element.childNodes[0]);
  }
}

/**
 * Returns an array of classes belonging to an element.
 */
function classList(element) {
  var retval = new Array();
  var possibleClasses = element.className.split(' ');
  for (var i in possibleClasses) {
    var className = possibleClasses[i];
    if (className.length) {
      retval[retval.length++] = className;
    }
  }
  return retval;
}

/**
 * Tests if an element is of a given class.
 */
function hasClass(element, className) {
  var classes = classList(element);
  for (var i in classes) {
    if (className == classes[i]) {
      return true;
    }
  }
  return false;
}

/**
 * Copies an object, and returns the copy.
 */
function copyObject(object) {
  var c = new Object();
  for (var i in object) {
    c[i] = object[i]
  }
  return c;
}

/**
 * Returns a string representation of a date, based on context.
 */
function formatDate(date) {
  var now = new Date();
  var dateString = "";
  if (!(date.getDate() == now.getDate() &&
        date.getMonth() == now.getMonth() &&
        date.getYear() == now.getYear())) {
    dateString += date.getFullYear() + "-" +
      ((date.getMonth() + 1) < 9 ? "0" : "") +
      (date.getMonth() + 1) + "-" +
      (date.getDate() < 10 ? "0" : "") +
      date.getDate() + " ";
  }
  dateString += date.getHours() + ":" +
    (date.getMinutes() < 10 ? "0" : "") +
    date.getMinutes() + ":" +
    (date.getSeconds() < 10 ? "0" : "") +
    date.getSeconds();
  
  return dateString;
}

/**
 * Use a file picker to allow the user to select an app binary.
 * @param app - the application's shortname (e.g., 'firefox')
 * @return - an nsIFile object of the selected app if the user pressed OK,
 *           otherwise undefined.
 * See: 
 * http://developer.mozilla.org/en/docs/XUL_Tutorial:Open_and_Save_Dialogs
 */
function chooseBinaryWithFilePicker(app) {
  var fp = getFilePicker();

  fp.init(window, "Open File", nsIFilePicker.modeOpen);
  fp.appendFilters(nsIFilePicker.filterApps);
  fp.appendFilters(nsIFilePicker.filterAll);

  var res = fp.show();
  if (res == nsIFilePicker.returnOK) {
    var file = fp.file;
    // TODO: show an error if the selected file doesn't match the expected
    // binary name?
    if (gPlatform == 'Macintosh') {
      // on mac, the file picker allows you only to pick an .app file, which
      // is really a directory, so we append to this the 'real' binary's
      // path
      file.append('Contents');
      file.append('MacOS');
      file.append(app + '-bin');
      if (!file.exists()) {
        throw(document.getElementById("bundle_profileManager").
                       getFormattedString('pathNotFoundError', [file.path]));
      }
    }
    return file;
  }
}

/**
 * Make all strings in an array the same length by right-padding them
 * with spaces.
 */
function equalizeLengths(strings) {
  var length = 0;
  for (var i in strings) {
    if (strings[i].length > length) {
      length = strings[i].length;
    }
  }
  var retval = new Array(strings.length);
  for (var i in strings) {
    retval[i] = strings[i];
    while (retval[i].length < length) {
      retval[i] = retval[i] + ' ';
    }
  }
  return retval;
}

/**
 * Display an alert or prompt to the user, and optionally allow the user to 
 * skip the alert in the future.
 * @param pref - the pref to check to see if the alert should be skipped,
 *               or null if the alert should never be skipped
 * @param title - the title to display in the alert dialog
 * @param message - the message to dispaly in the alert dialog
 * @param confirm - if true, the alert will be a confirmation prompt, with
 *                  OK and cancel buttons
 * @return true if the user selected OK, or the alert was not displayed
 *         due to a user preference, otherwise false
 */
function doAlertCheck(pref, title, message, confirm) {
  var returnval = true;
  var skipalert = false;
  try {
    if (pref && prefService.getBoolPref(pref))
      skipalert = true;
  }
  catch (e) { }
  if (!skipalert) {
    var bundle = document.getElementById("bundle_profileManager");
    var result = {value: false};
    if (confirm) {
      if (pref) {
        returnval = promptService.confirmCheck(
          window,
          title,
          message,
          bundle.getString('dontShowPromptAgainText'), 
          result);
      }
      else {
        returnval = promptService.confirm(window, title, message);
      }
    }
    else {
      if (pref) {
        promptService.alertCheck(window,
          title,
          message,
          bundle.getString('dontShowMessageAgainText'),
          result);
      }
      else {
        promptService.alert(window, title, message);
      }
    }
    if (pref && result.value) {
      prefService.setBoolPref(pref, true);
    }
  }

  return returnval;
}

/**
 * Crops a string that presumptively begins with a path, until it fits
 * into a specified width.  This is done by iteratively truncating parts
 * of the path, from the left, and replacing the truncated portion of the
 * string with an ellipsis.
 * @param text - the text that needs cropping
 * @param width - the width, in pixels, that the cropped text must fit in
 * @param ctx - a canvas 2d context, that we can use for measuring text
 * @return the cropped string
 */
function cropPathString(text, width, ctx) {
  var pathSep = (gPlatform == "Windows" ? '\\' : '/');
  var parts = text.split(pathSep);
  for (var i = 1; i < parts.length; i++) {
    let newtext = "..." + parts.slice(i).join(pathSep);
    if (ctx.measureText(newtext).width <= width)
      return newtext;
  }
  // we should never realistically get here, but just to be safe:
  return parts[parts.length-1];
}
